/***************************************************************************
 *   Copyright (C) 2007  Antony Martin <martin.antony@yahoo.fr>            *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or         *
 *   modify it under the terms of the GNU General Public License           *
 *   as published by the Free Software Foundation; either version 2        *
 *   of the License, or (at your option) any later version.                *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA          *
 ***************************************************************************/

/* Cree le : 16 mai 2007
   derniere modification : 27/06/2007 */


#ifndef __SCEMODEL_H__
#define __SCEMODEL_H__


/* constantes propres a Model */
/* nombre max de niveaux de detail geometrique */
#define SCE_MAX_GEOMETRY_LOD 8


typedef struct sce_smodel SCE_SModel;
struct sce_smodel
{
    /* medias */
    /* maillage(s) du modele */
    SCE_SMesh *mesh[SCE_MAX_GEOMETRY_LOD];
    SCE_SMaterial *mat;     /* materiau du modele */
    
    unsigned int mesh_lod;  /* mesh actif pour le rendu */
    
    /* mouvements */
    SCE_TVector3 pos;       /* position du modele dans la scene */
    
    SCE_SInertVar angle;    /* angle de rotation */
    SCE_TVector3 axe;       /* axe de rotation */
    
    /* angles de rotation (bonus) */
    SCE_SInertVar angle_x;
    SCE_SInertVar angle_y;
    SCE_SInertVar angle_z;
    
    
    /* echelle du modele */
    SCE_SInertVar scale_x;
    SCE_SInertVar scale_y;
    SCE_SInertVar scale_z;
    
    SCE_TVector3 front;     /* axe frontal, les deplacements lineaires
                               seront effectues a base de celui-ci. il
                               represente "l'avant" du modele */
    int affect_front;       /* booleen, true: le vecteur ci-dessus
                               est affecte par les rotations */
    
    SCE_SInertVar front_mv;     /* mouvements/deplacements frontaux */
    
    
    int animated;           /* booleen, true: modele anime */
};


/* initialise le gestionnaire de modeles */
int SCE_Model_Init(void);

/* quitte le gestionnaire de modeles */
int SCE_Model_Quit(void);


/* cree un modele selon un critere: avec ou sans animation possible
   (impossible de revenir sur un choix a part supprimer le modele) */
SCE_SModel* SCE_Model_Create(const int);

/* supprime un modele */
void SCE_Model_Delete(SCE_SModel*);


/* positionne le model a un endroit precis */
void SCE_Model_SetPos(SCE_SModel*, float, float, float);
void SCE_Model_SetPosv(SCE_SModel*, SCE_TVector3);

/* defini l'axe de rotation du modele sur lui-meme */
void SCE_Model_SetAxe(SCE_SModel*, float, float, float);
void SCE_Model_SetAxev(SCE_SModel*, SCE_TVector3);

/* defini les angles de rotation du modele */
void SCE_Model_Rotate(SCE_SModel*, float, float, float, float);

/* defini si le vecteur frontal du modele sera affecte par les rotations */
void SCE_Model_AffectFrontVec(SCE_SModel*, int);


/* calcul les translations et les rotations que doit effectuer le modele */
void SCE_Model_ComputeMovs(SCE_SModel*);


/* configure les matrices pour un modele */
void SCE_Model_PushMatrices(SCE_SModel*);

/* restaure les matrices modifiees */
void SCE_Model_PopMatrices(void);


/* bind le modele entier */
void SCE_Model_Bind(SCE_SModel*);

/* bind le mesh du modele */
void SCE_Model_BindMesh(SCE_SModel*);

/* bind le materiau du modele */
void SCE_Model_BindMaterial(SCE_SModel*);

/* dessine le modele binde */
void SCE_Model_Draw(void);


/* effectue le rendu d'un modele */
void SCE_Model_Render(SCE_SModel*);


/* charge un mesh (last param = use LOD) */
int SCE_Model_LoadMesh(SCE_SModel*, const char*, int);


/* supprime un mesh */
void SCE_Model_DeleteLOD(SCE_SModel*, SCEuint);


#endif /* guard */
