/***************************************************************************
 *   Copyright (C) 2007  Antony Martin <martin.antony@yahoo.fr>            *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or         *
 *   modify it under the terms of the GNU General Public License           *
 *   as published by the Free Software Foundation; either version 2        *
 *   of the License, or (at your option) any later version.                *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA          *
 ***************************************************************************/

/* Cree le : 21 decembre 2006
   derniere modification le 27/06/2007 */


#ifndef __SCEVECTOR_H__
#define __SCEVECTOR_H__


typedef float SCE_TVector4[4];

typedef float SCE_TVector3[3];

typedef float SCE_TVector2[2];

/* 3D retourne la distance qui separe le point a du point b */
#define SCE_Vector3_Dst(a, b) SCE_Math_Sqrt((((a)[0]-(b)[0])*((a)[0]-(b)[0]))+\
                                            (((a)[1]-(b)[1])*((a)[1]-(b)[1]))+\
                                            (((a)[2]-(b)[2])*((a)[2]-(b)[2])))
/* 2D */
#define SCE_Vector2_Dst(a, b) SCE_Math_Sqrt((((a)[0]-(b)[0])*((a)[0]-(b)[0]))+\
                                            (((a)[1]-(b)[1])*((a)[1]-(b)[1])))


/* 3D retourne la longueur du vecteur a (par rapport au point (0,0,0) ) */
#define SCE_Vector3_Lenght(a) SCE_Math_Sqrt(((a)[0]*(a)[0])+\
                                            ((a)[1]*(a)[1])+\
                                            ((a)[2]*(a)[2]))
/* 2D */
#define SCE_Vector2_Lenght(a) SCE_Math_Sqrt(((a)[0]*(a)[0])+((a)[1]*(a)[1]))



/* 3D deplace le vecteur v dans la direction
      pointee par d multiplie par la longueur l */
#define SCE_Vector3_Translate(v, d, l) (v)[0] += (d)[0]*(l);\
                                       (v)[1] += (d)[1]*(l);\
                                       (v)[2] += (d)[2]*(l)
/* 2D */
#define SCE_Vector2_Translate(v, d, l) (v)[0] += (d)[0]*(l);\
                                       (v)[1] += (d)[1]*(l)


/* 3D additionne deux vecteurs, le resultat est stocke dans v1 */
#define SCE_Vector3_Addv(v1, v2, v3) (v1)[0] = (v2)[0]+(v3)[0];\
                                     (v1)[1] = (v2)[1]+(v3)[1];\
                                     (v1)[2] = (v2)[2]+(v3)[2]
/* 2D */
#define SCE_Vector2_Addv(v1, v2, v3) (v1)[0] = (v2)[0]+(v3)[0];\
                                     (v1)[1] = (v2)[1]+(v3)[1]
/* version 'numerique' */
#define SCE_Vector3_Add(v1, v2, n) (v1)[0] = (v2)[0]+(n);\
                                   (v1)[1] = (v2)[1]+(n);\
                                   (v1)[2] = (v2)[2]+(n)
/* 2D */
#define SCE_Vector2_Add(v1, v2, n) (v1)[0] = (v2)[0]+(n);\
                                   (v1)[1] = (v2)[1]+(n)


/* 3D soustrait deux vecteurs, le resultat est stocke dans v1 */
#define SCE_Vector3_Subv(v1, v2, v3) (v1)[0] = (v2)[0]-(v3)[0];\
                                     (v1)[1] = (v2)[1]-(v3)[1];\
                                     (v1)[2] = (v2)[2]-(v3)[2]
/* 2D */
#define SCE_Vector2_Subv(v1, v2, v3) (v1)[0] = (v2)[0]-(v3)[0];\
                                     (v1)[1] = (v2)[1]-(v3)[1]
/* version 'numerique' */
#define SCE_Vector3_Sub(v1, v2, n) (v1)[0] = (v2)[0]-(n);\
                                   (v1)[1] = (v2)[1]-(n);\
                                   (v1)[2] = (v2)[2]-(n)
/* 2D */
#define SCE_Vector2_Sub(v1, v2, n) (v1)[0] = (v2)[0]-(n);\
                                   (v1)[1] = (v2)[1]-(n)


/* 3D multiplie deux vecteurs, le resultat est stocke dans v1 */
#define SCE_Vector3_Prodv(v1, v2, v3) (v1)[0] = (v2)[0]*(v3)[0];\
                                      (v1)[1] = (v2)[1]*(v3)[1];\
                                      (v1)[2] = (v2)[2]*(v3)[2]
/* 2D */
#define SCE_Vector2_Prodv(v1, v2, v3) (v1)[0] = (v2)[0]*(v3)[0];\
                                      (v1)[1] = (v2)[1]*(v3)[1]
/* version 'numerique' */
#define SCE_Vector3_Prod(v1, v2, n) (v1)[0] = (v2)[0]*(n);\
                                    (v1)[1] = (v2)[1]*(n);\
                                    (v1)[2] = (v2)[2]*(n)
/* 2D */
#define SCE_Vector2_Prod(v1, v2, n) (v1)[0] = (v2)[0]*(n);\
                                    (v1)[1] = (v2)[1]*(n)


/* 3D divise deux vecteurs, le resultat est stocke dans v1 */
#define SCE_Vector3_Divv(v1, v2, v3) (v1)[0] = (v2)[0]/(v3)[0];\
                                     (v1)[1] = (v2)[1]/(v3)[1];\
                                     (v1)[2] = (v2)[2]/(v3)[2]
/* 2D */
#define SCE_Vector2_Divv(v1, v2, v3) (v1)[0] = (v2)[0]/(v3)[0];\
                                     (v1)[1] = (v2)[1]/(v3)[1]
/* version 'numerique' */
#define SCE_Vector3_Div(v1, v2, n) (v1)[0] = (v2)[0]/(n);\
                                   (v1)[1] = (v2)[1]/(n);\
                                   (v1)[2] = (v2)[2]/(n)
/* 2D */
#define SCE_Vector2_Div(v1, v2, n) (v1)[0] = (v2)[0]/(n);\
                                   (v1)[1] = (v2)[1]/(n)


/* 3D place toutes les composantes d'un vecteur a la valeur specifiee par n */
#define SCE_Vector3_Init(v, n) (v)[0] = n;\
                               (v)[1] = n;\
                               (v)[2] = n
/* 2D */
#define SCE_Vector2_Init(v, n) (v)[0] = n;\
                               (v)[1] = n



/* 3D copie v2 dans v1 */
#define SCE_Vector3_Affectv(v1, v2) (v1)[0] = (v2)[0];\
                                    (v1)[1] = (v2)[1];\
                                    (v1)[2] = (v2)[2]
/* 2D */
#define SCE_Vector2_Affectv(v1, v2) (v1)[0] = (v2)[0];\
                                    (v1)[1] = (v2)[1]
/* version 'numerique' */
#define SCE_Vector3_Affect(v1, a, b, c) (v1)[0] = a;\
                                        (v1)[1] = b;\
                                        (v1)[2] = c
/* 2D */
#define SCE_Vector2_Affect(v1, a, b) (v1)[0] = a;\
                                     (v1)[1] = b



/* 3D produit scalaire */
#define SCE_Vector3_Dot(v1, v2) (((v1)[0]*(v2)[0])+\
                                 ((v1)[1]*(v2)[1])+\
                                 ((v1)[2]*(v2)[2]))
/* 2D */
#define SCE_Vector2_Dot(v1, v2) (((v1)[0]*(v2)[0])+\
                                 ((v1)[1]*(v2)[1]))



/* 3D produit vectoriel */
#define SCE_Vector3_Cross(v1, v2, v3) (v1)[0] =\
                                            (v2)[1]*(v3)[2] - (v2)[2]*(v3)[1];\
                                      (v1)[1] =\
                                            (v2)[2]*(v3)[0] - (v2)[0]*(v3)[2];\
                                      (v1)[2] =\
                                            (v2)[0]*(v3)[1] - (v2)[1]*(v3)[0]


/* 3D operateur */
#define SCE_Vector3_Operator1(v, o, n)\
    (v)[0] o n; (v)[1] o n; (v)[2] o n

#define SCE_Vector3_Operator1v(v1, o, v2)\
    (v1)[0] o (v2)[0];\
    (v1)[1] o (v2)[1];\
    (v1)[2] o (v2)[2]

#define SCE_Vector3_Operator2(v, o1, v2, o2, n)\
    (v)[0] o1 (v2)[0] o2 n;\
    (v)[1] o1 (v2)[1] o2 n;\
    (v)[2] o1 (v2)[2] o2 n

#define SCE_Vector3_Operator2v(v1, o1, v2, o2, v3)\
    (v1)[0] o1 (v2)[0] o2 (v3)[0];\
    (v1)[1] o1 (v2)[1] o2 (v3)[1];\
    (v1)[2] o1 (v2)[2] o2 (v3)[2]

#define SCE_Vector3_Operator3(v1, o1, v2, o2, v3, o3, n)\
    (v1)[0] o1 (v2)[0] o2 (v3)[0] o3 n;\
    (v1)[1] o1 (v2)[1] o2 (v3)[1] o3 n;\
    (v1)[2] o1 (v2)[2] o2 (v3)[2] o3 n

#define SCE_Vector3_Operator3v(v1, o1, v2, o2, v3, o3, v4)\
    (v1)[0] o1 (v2)[0] o2 (v3)[0] o3 (v4)[0];\
    (v1)[1] o1 (v2)[1] o2 (v3)[1] o3 (v4)[1];\
    (v1)[2] o1 (v2)[2] o2 (v3)[2] o3 (v4)[2]

/* 2D */
#define SCE_Vector2_Operator1(v, o, n) (v)[0] o n; (v)[1] o n
#define SCE_Vector2_Operator1v(v1, o, v2) (v1)[0] o (v2)[0];\
                                         (v1)[1] o (v2)[1]

#define SCE_Vector2_Operator2(v, o1, v2, o2, n) (v)[0] o1 (v2)[0] o2 n;\
                                                (v)[1] o1 (v2)[1] o2 n

#define SCE_Vector2_Operator2v(v1, o1, v2, o2, v3) (v1)[0] o1 (v2)[0] o2 (v3)[0];\
                                                  (v1)[1] o1 (v2)[1] o2 (v3)[1]

#define SCE_Vector2_Operator3v(v1, o1, v2, o2, v3, o3, v4)\
    (v1)[0] o1 (v2)[0] o2 (v3)[0] o3 (v4)[0];\
    (v1)[1] o1 (v2)[1] o2 (v3)[1] o3 (v4)[1]


/* 3D normalise un vecteur */
void SCE_Vector3_Normalize(SCE_TVector3);
/* 2D */
void SCE_Vector2_Normalize(SCE_TVector2);


/* effectue une rotation (axe X) a un vecteur */
void SCE_Vector3_RotateX(SCE_TVector3, float, float);

/* effectue une rotation (axe Y) a un vecteur */
void SCE_Vector3_RotateY(SCE_TVector3, float, float);

/* effectue une rotation (axe Z) a un vecteur */
void SCE_Vector3_RotateZ(SCE_TVector3, float, float);



#endif /* guard */
