/***************************************************************************
 *   Copyright (C) 2007  Antony Martin <martin.antony@yahoo.fr>            *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or         *
 *   modify it under the terms of the GNU General Public License           *
 *   as published by the Free Software Foundation; either version 2        *
 *   of the License, or (at your option) any later version.                *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA          *
 ***************************************************************************/

/* Cree le : 19 janvier 2007
   derniere modification le 25/06/2007 */


/**

IMPORTANT!!:
______
\p---> separer la conversion de masse de variables de types differents
       du generateur TBN, l'implementer dans une fonction a part
       (SCEConvert.c ??) ---> a faire... (18/05/2007 : en cours ...
                                          19/05/2007 : ca sera pas fait, trop
                                            specifique dans le cas des meshs,
                                            trop complique, trop chiant, trop
                                            con)

\p---> implémenter la suppression des données des tampons (si demande).
\p---> programmer un generateur de LOD (on verra quand j'aurai que ca a foutre)

**/


#include <SCE/SCEMinimal.h>

#include <SCE/utils/SCEVector.h>
#include <SCE/utils/SCEString.h>

#include <SCE/core/SCECBuffers.h>

#include <SCE/interface/SCEMesh.h>



/* mesh en cache ('binde') */
static SCE_SMesh *m;


int SCE_Mesh_Init(void)
{
    return (SCE_OK);
}

int SCE_Mesh_Quit(void)
{
    return (SCE_OK);
}


SCE_SMesh* SCE_Mesh_Create(void)
{
    SCE_SMesh *mesh = NULL;
    unsigned int i;
    
    mesh = SCE_malloc(sizeof *mesh);
    if(mesh == NULL)
    {
        Logger_LogSrc();
        return (NULL);
    }
    
    mesh->n_active_vb = 0;
    mesh->active_vcount = 0;
    mesh->active_icount = 0;
    mesh->builded = 0;
    mesh->use_indices = 0;
    mesh->ib_active = 0;
    mesh->n_vb = 0;
    mesh->n_ib = 0;
    mesh->polygon_type = GL_POINTS; /* evitons tout probleme */
    
    /* NOTE: devient errone si le nombre de tampon passe en illimite */
    for(i=0; i<SCE_MAX_BUFFERS; i++)
    {
        mesh->vb_actives[i] = 0;
        
        mesh->vb[i].builded = 0;
        mesh->vb[i].count = 0;
        mesh->vb[i].buffer = NULL;
        /* mode par defaut des tampons de sommets */
        mesh->vb[i].usage = GL_STREAM_DRAW;
        
        
        mesh->ib[i].builded = 0;
        mesh->ib[i].count = 0;
        mesh->ib[i].buffer = NULL;
        mesh->ib[i].data_type = SCE_UNSIGNED_BYTE;
        /* mode par defaut des tampons d'indices */
        mesh->ib[i].usage = GL_STATIC_DRAW;
    }
    
    return (mesh);
}


int SCE_Mesh_Delete(SCE_SMesh *mesh)
{
    unsigned int i;
    
    if(mesh != NULL)
    {
        for(i=0; i<SCE_MAX_BUFFERS; i++)
        {
            SCE_CDeleteVertexBuffer(mesh->vb[i].buffer);
            SCE_CDeleteIndexBuffer(mesh->ib[i].buffer);
        }
        
        if(SCE_free(mesh) < 0)
        {
            Logger_LogSrc();
            return (SCE_ERROR);
        }
    }
    
    return (SCE_OK);
}


int SCE_Mesh_AddVB(SCE_SMesh *mesh, SCEenum usage)
{
    #ifdef __SCE_DEBUG__
    if(mesh == NULL)
    {
        Logger_Log(SCE_INVALID_ARG);
        return (SCE_ERROR);
    }
    
    if(mesh->n_vb >= SCE_MAX_BUFFERS)
    {
        Logger_Log(SCE_INVALID_OPERATION);
        Logger_LogMsg("you can't add a new buffer, mesh is at maximum buffers");
        return (SCE_ERROR);
    }
    
    if(usage != GL_STATIC_DRAW &&
       usage != GL_STREAM_DRAW &&
       usage != GL_DYNAMIC_DRAW &&
       usage != 0) /* 0 ne precise aucune valeur et laisse celle par defaut */
    {
        Logger_Log(SCE_INVALID_ARG);
        Logger_LogMsg("unknown buffer usage");
        return (SCE_ERROR);
    }
    #endif
    
    if(usage != 0)
        mesh->vb[mesh->n_vb].usage = usage;
    
    mesh->vb[mesh->n_vb].buffer = SCE_CCreateVertexBuffer();
    if(mesh->vb[mesh->n_vb].buffer == NULL)
    {
        Logger_LogSrc();
        return (SCE_ERROR);
    }
    
    mesh->n_vb++;
    
    /* on retourne l'identifiant du tampon que l'on vient de creer */
    return (mesh->n_vb-1);
}

void SCE_Mesh_DeleteVB(SCE_SMesh *mesh, unsigned int id)
{
    if(mesh != NULL)
    {
        SCE_CDeleteVertexBuffer(mesh->vb[id].buffer);
        
        mesh->vb[id].builded = 0;
        mesh->vb[id].count = 0;
        mesh->vb[id].buffer = NULL;
        /* mode par defaut des tampons de sommets */
        mesh->vb[id].usage = GL_STREAM_DRAW;
        
        /* si le tampon etait actif pour le rendu */
        if(mesh->vb_actives[id] != 0)
        {
            /* on le desactive, et on desincremente le compteur de tampons */
            mesh->vb_actives[id] = 0;
            mesh->n_active_vb--;
        }
    }
}


int SCE_Mesh_AddIB(SCE_SMesh *mesh, SCEenum usage, SCEenum data_type)
{
    #ifdef __SCE_DEBUG__
    if(mesh == NULL)
    {
        Logger_Log(SCE_INVALID_ARG);
        return (SCE_ERROR);
    }
    
    if(mesh->n_ib >= SCE_MAX_BUFFERS)
    {
        Logger_Log(SCE_INVALID_OPERATION);
        Logger_LogMsg("you can't add a new buffer, mesh is at maximum buffers");
        return (SCE_ERROR);
    }
    
    if(data_type != SCE_INT &&
       data_type != SCE_UNSIGNED_INT &&
       data_type != SCE_SHORT &&
       data_type != SCE_UNSIGNED_SHORT &&
       data_type != SCE_BYTE &&
       data_type != SCE_UNSIGNED_BYTE)
    {
        Logger_Log(SCE_INVALID_ARG);
        Logger_LogMsg("unknown data type or type not supported");
        return (SCE_ERROR);
    }
    
    if(usage != GL_STATIC_DRAW &&
       usage != GL_STREAM_DRAW &&
       usage != GL_DYNAMIC_DRAW &&
       usage != 0) /* 0 ne precise aucune valeur et laisse celle par defaut */
    {
        Logger_Log(SCE_INVALID_ARG);
        Logger_LogMsg("unknown buffer usage");
        return (SCE_ERROR);
    }
    #endif
    
    if(usage != 0)
        mesh->ib[mesh->n_ib].usage = usage;
    
    mesh->ib[mesh->n_ib].data_type = data_type;
    
    mesh->ib[mesh->n_ib].buffer = SCE_CCreateIndexBuffer();
    if(mesh->ib[mesh->n_ib].buffer == NULL)
    {
        Logger_LogSrc();
        return (SCE_ERROR);
    }
    
    mesh->n_ib++;
    
    /* on retourne l'identifiant du tampon que l'on vient de creer */
    return (mesh->n_ib-1);
}

void SCE_Mesh_DeleteIB(SCE_SMesh *mesh, unsigned int id)
{
    if(mesh != NULL)
    {
        SCE_CDeleteIndexBuffer(mesh->ib[id].buffer);
        
        mesh->ib[id].builded = 0;
        mesh->ib[id].count = 0;
        mesh->ib[id].buffer = NULL;
        mesh->ib[id].data_type = SCE_UNSIGNED_BYTE;
        /* mode par defaut des tampons d'indices */
        mesh->ib[id].usage = GL_STATIC_DRAW;
        
        /* si le tampon d'indices etait utilise */
        if(mesh->ib_active == id)
        {
            /* on desactive les indices */
            mesh->use_indices = 0;
            /* remise a 0 du count d'indices par defaut du mesh */
            mesh->active_icount = 0;
        }
    }
}


/* revise le 10/07/2007 */
int SCE_Mesh_ActiveVB(SCE_SMesh *mesh, unsigned int buffer_id, int active)
{
    #ifdef __SCE_DEBUG__
    if(buffer_id >= SCE_MAX_BUFFERS)
    {
        Logger_Log(SCE_INVALID_ARG);
        Logger_LogMsg("buffer id is too long, the maximum is %d",
                       SCE_MAX_BUFFERS-1);
        return (SCE_ERROR);
    }
    #endif
    
    if(mesh == NULL)
    {
        /* alors on va travailler sur le mesh actif pour le rendu (m) en
           faisant bien attention a ne pas changer ses etats! (vb_active) */
        
        /* il faut bien garder en tete que cette technique est vraiment du
           brut-force dans la mesure ou tous les vertex buffers n'ont pas
           le meme count, et puisqu'il s'agit d'un rendu dynamique, on ne
           verifie pas cette eventuelle incompatibilite, donc, danger... */
        if(m && active)
            SCE_CUseVertexBuffer(m->vb[buffer_id].buffer);
    }
    else
    {
        /* sinon, changement des etats du mesh donne en argument.
           mais surtout, il faut verifier son count avec celui deja present */
        
        #ifdef __SCE_DEBUG__
        /* on verifie si le mesh est construit
           (sinon le programme plante) */
        if(!mesh->builded)
        {
            Logger_Log(SCE_INVALID_OPERATION);
            Logger_LogMsg("you can't activate any buffers"
                          " of a non-builded mesh");
            return (SCE_ERROR);
        }
        #endif
        
        /* on verifie si l'activation est demandee... */
        if(active)
        {
            /* ...et si le tampon n'est pas deja actif pour le rendu */
            if(!mesh->vb_actives[buffer_id])
            {
                /* verification du count */
                
                if(mesh->active_vcount == 0)
                {
                    /* aucun count actif, on le cree */
                    mesh->active_vcount = mesh->vb[buffer_id].count;
                }
                
                /* s'il est egal a celui deja present */
                if(mesh->active_vcount == mesh->vb[buffer_id].count)
                {
                    /* alors on peut activer le tampon */
                    mesh->vb_actives[buffer_id] = 1;
                    mesh->n_active_vb++;
                }
                #ifdef __SCE_DEBUG__
                /* mode debug -> sinon, on genere une erreur */
                else
                {
                    Logger_Log(SCE_INVALID_OPERATION);
                    Logger_LogMsg("you can't activate a buffer with a count not"
                                  " equal of the active default mesh count");
                    return (SCE_ERROR);
                }
                #endif
            }
        }
        else
        {
            /* si le tampon etait actif, on desincremente le compteur
               de tampons actifs */
            if(mesh->vb_actives[buffer_id])
                mesh->n_active_vb--;
            
            mesh->vb_actives[buffer_id] = 0;
            
            /* reinitialisation du vcount par defaut du mesh si plus aucun
               vertex buffer n'est actif pour le rendu */
            if(mesh->n_active_vb == 0)
                mesh->active_vcount = 0;
        }
    }
    
    return (SCE_OK);
}

/* revise le 10/07/2007 */
int SCE_Mesh_ActiveIB(SCE_SMesh *mesh, unsigned int buffer_id, int active)
{
    /* ca fait parti du debuggage ca non.. ? */
    #ifdef __SCE_DEBUG__
    /* NOTE: devient errone si le nombre de tampons devient illimite */
    if(buffer_id >= SCE_MAX_BUFFERS)
    {
        Logger_Log(SCE_INVALID_ARG);
        Logger_LogMsg("buffer id is too long, the maximum is %d",
                       SCE_MAX_BUFFERS-1);
        return (SCE_ERROR);
    }
    #endif
    
    if(mesh == NULL)
    {
        /* alors on va travailler sur le mesh actif pour le rendu (m)
           en faisant bien attention a ne pas changer ses etats!
           (use_indices, ib_active) */
        
        
        /* brut-force dans la mesure ou le count du tampon d'indices ci-present
           doit etre egal au count actif par defaut du mesh, sinon, on s'expose
           a de graves ennuis... (seg faults et autres) */
        if(m)
        {
            if(active)
                SCE_CUseIndexBuffer(m->ib[buffer_id].buffer);
            else
                SCE_CUseIndexBuffer(NULL);
        }
    }
    else
    {
        /* sinon, changement des etats du mesh donne en argument */
        
        #ifdef __SCE_DEBUG__
        /* on verifie si le mesh est construit
           (sinon le programme plante) */
        if(!mesh->builded)
        {
            Logger_Log(SCE_INVALID_OPERATION);
            Logger_LogMsg("you can't activate any buffers"
                          " of a non-builded mesh");
            return (SCE_ERROR);
        }
        if(mesh->ib[buffer_id].buffer == NULL)
        {
            /* tentative d'activation d'un tampon non initialise... */
            Logger_Log(SCE_INVALID_OPERATION);
            Logger_LogMsg("you can't activate a uninitialized buffer");
            return (SCE_ERROR);
        }
        if(!mesh->ib[buffer_id].builded)
        {
            /* tentative d'activation d'un tampon non construit... */
            Logger_Log(SCE_INVALID_OPERATION);
            Logger_LogMsg("you can't activate an unbuilded buffer");
            return (SCE_ERROR);
        }
        #endif
        
        if(!active)
            mesh->use_indices = 0;
        else
        {
            mesh->use_indices = 1;
            mesh->ib_active = buffer_id;
            mesh->active_icount = mesh->ib[buffer_id].count;
        }
    }
    
    return (SCE_OK);
}


/* recherche le tampon qui contient les mêmes
   donnees que celles envoyees en parametre */
static int SCE_Mesh_FindData(SCE_SMeshVertexBuffer *vb, void *data)
{
    unsigned int i;
    
    for(i=0; i<vb->buffer->n_data; i++)
    {
        if(vb->buffer->datas[i].data == data)
            return (i);
    }
    
    return (SCE_ERROR);
}



int
SCE_Mesh_AddVertices(
  SCE_SMesh *mesh,          /* mesh sur lequel appliquer les modifications */
  unsigned int id,          /* tampon dans lequel placer les donnees */
  unsigned int type,        /* type : SCE_POSITION, SCE_COLOR, ... */
  SCEenum data_type,        /* type des donnees: GL_FLOAT, GL_INT, ... */
  unsigned int data_nvpv,   /* number of variables per vertex */
  size_t data_size,         /* nombre de donnees (nombre total de variables) */
  void *data,               /* donnees */
  int free_data)            /* autorisation de supprimmer les donnees */
{
    /* taille d'une variable des donnees */
    size_t var_size;
    
    /* flottant de verification des tailles */
    float count_verif;
    
    /* identifiant du tampon des donnees si elles sont doubles */
    int double_id;
    
    /* identifiant du declarateur de vertex actif dans le tampon mis a jour */
    unsigned int decid;
    
    
    /* verifications en plusieurs fois:
       pour localiser avec precision le parametre errone */
    
    if(data == NULL)
    {
        #ifdef __SCE_DEBUG__
        Logger_Log(SCE_INVALID_ARG);
        return (SCE_ERROR);
        #else
        return (SCE_OK);
        #endif
    }
    
    if(type == SCE_NORMAL)
        data_nvpv = 3;
    
    #ifdef __SCE_DEBUG__
    if(mesh == NULL)
    {
        Logger_Log(SCE_INVALID_ARG);
        return (SCE_ERROR);
    }
    if((data_nvpv < 1 || data_nvpv > 4) && type < SCE_TEXCOORD0)
    {
        Logger_Log(SCE_INVALID_ARG);
        return (SCE_ERROR);
    }
    if(type == SCE_COLOR && data_nvpv < 3)
    {
        Logger_Log(SCE_INVALID_ARG);
        return (SCE_ERROR);
    }
    if(type == SCE_POSITION &&
        (data_type == SCE_BYTE || data_type == SCE_UNSIGNED_BYTE ||
         data_type == SCE_UNSIGNED_SHORT || data_type == SCE_UNSIGNED_INT))
    {
        Logger_Log(SCE_INVALID_ARG);
        return (SCE_ERROR);
    }
    if(type == SCE_NORMAL &&
        (data_type == SCE_UNSIGNED_BYTE || data_type == SCE_UNSIGNED_SHORT ||
         data_type == SCE_UNSIGNED_INT))
    {
        Logger_Log(SCE_INVALID_ARG);
        return (SCE_ERROR);
    }
    /* NOTE: devient errone si le nombre de tampon passe en illimite */
    if(id >= SCE_MAX_BUFFERS)
    {
        Logger_Log(SCE_INVALID_ARG);
        return (SCE_ERROR);
    }
    
    if(mesh->vb[id].buffer == NULL)
    {
        Logger_Log(SCE_INVALID_OPERATION);
        Logger_LogMsg("the buffer %d of specified mesh is not initialized",
                      id);
        return (SCE_ERROR);
    }
    #endif
    /***/
    
    
    var_size = SCE_CSizeofType(data_type);
    if(var_size <= 0)
    {
        Logger_LogSrc();
        return (SCE_ERROR);
    }
    
    /* nombre_variables / variables_par_vertex = nombre de vertices
       si le resultat est entier, OK */
    count_verif = (float)data_size/(float)data_nvpv;
    
    /* verification de la taille des donnees */
    if( (count_verif - (size_t)count_verif) != 0 )
    {
        /* -> count flottant, erreur */
        Logger_Log(SCE_INVALID_ARG);
        Logger_LogMsg("invalid data size(s), final vertex count is a float!");
        return (SCE_ERROR);
    }
    else if(mesh->vb[id].count == 0)
    {
        /* aucun 'count' de vertex n'a encore ete envoye, alors on le fait */
        mesh->vb[id].count = data_size/data_nvpv;
    }
    else if(mesh->vb[id].count != data_size/data_nvpv)
    {
        /* le 'count' deja present differe de celui calcule */
        Logger_Log(SCE_INVALID_ARG);
        Logger_LogMsg("invalid data size: not equal to current");
        return (SCE_ERROR);
    }
    
    /* affectation de l'id du declarateur de vertex */
    decid = mesh->vb[id].buffer->n_decs;
    
    /* reallocation des declarations de vertex, on en ajoute un */
    mesh->vb[id].buffer->dec =
    SCE_realloc( mesh->vb[id].buffer->dec,
                 (decid+1) * sizeof(*mesh->vb[id].buffer->dec) );
    if(mesh->vb[id].buffer->dec == NULL) /** NOTE: il faut que realloc renvoie NULL en cas d'erreur */
    {
        Logger_LogSrc();
        return (SCE_ERROR);
    }
    
    /* stockage des info dans le declarateur de vertex */
    mesh->vb[id].buffer->dec[decid].v_type = type;
    mesh->vb[id].buffer->dec[decid].type = data_type;
    mesh->vb[id].buffer->dec[decid].n_vars = data_nvpv;
    /* NOTE: stride n'est pas modifiable, mais il devrait l'etre par la suite */
    mesh->vb[id].buffer->dec[decid].stride = 0;
    
    /* recuperation de la zone des donnees doubles */
    double_id = SCE_Mesh_FindData(&mesh->vb[id], data);
    if(double_id > -1)
    {
        /* si les donnees sont des donnees double (deja stockee)
           on recupere la zone ou sont stockees ces donnees
           pour la designer comme zone de donnees au declarateur */
        mesh->vb[id].buffer->dec[decid].first =
        mesh->vb[id].buffer->datas[double_id].first;
    }
    else
    {
        /* sinon on recupere la zone suivante
           de stockage de donnees dans le tampon */
        mesh->vb[id].buffer->dec[decid].first =
        SCE_CGetVertexBufferFirst(mesh->vb[id].buffer);
        
        /* et on envoie les donnees a cette zone */
        SCE_CSetVertexBufferData(mesh->vb[id].buffer, free_data,
                                 data_size*var_size, data);
    }
    
    /* + une declaration de vertex */
    mesh->vb[id].buffer->n_decs++;
    
    /* ...donc le tampon est a reconstruire pour etre valide */
    mesh->vb[id].builded = 0;
    
    return (SCE_OK);
}


/*
int
SCE_Mesh_Fill(
  SCE_SMesh *mesh,
  unsigned int id, (buffer_id)
  unsigned int n_decs,
  SCE_CVertexDeclaration *dec,
  size_t vertex_size,
  unsigned int n_vertices,
  void *data)
{
    if(mesh == NULL || dec == NULL)
    {
        Logger_Log(SCE_INVALID_ARG);
        return (-1);
    }
    
    
    
    return (SCE_OK);
}*/


int
SCE_Mesh_AddIndices(
  SCE_SMesh *mesh,      /* mesh a remplir */
  unsigned int id,      /* ID du tampon d'indices ou stocker les donnees */
  size_t i_count,       /* nombre de donnees */
  void *indices,        /* pointeur sur les donnees */
  int free_data)        /* autorisation de supprimmer les donnees */
{
    #ifdef __SCE_DEBUG__
    if(mesh == NULL)
    {
        Logger_Log(SCE_INVALID_ARG);
        return (SCE_ERROR);
    }
    #endif
    
    if(mesh->ib[id].buffer == NULL)
    {
        Logger_Log(SCE_INVALID_OPERATION);
        Logger_LogMsg("the buffer %d of the specified mesh is not initialized",
                      id);
        return (SCE_ERROR);
    }
    
    SCE_CSetIndexBufferData(mesh->ib[id].buffer, free_data,
                            i_count*SCE_CSizeofType(mesh->ib[id].data_type),
                            indices);
    
    mesh->ib[id].count = i_count;
    
    /* le tampon est a reconstruire pour etre valide */
    mesh->ib[id].builded = 0;
    
    return (SCE_OK);
}


/* calcul les tangentes, normales et binormales d'un triangle.
   index represente un tableau de 3 indices qui seront utilises
   pour dereferencer les tableaux (tous)
   les tableaux tangents, binormals et normals peuvent etres NULL,
   dans quel cas ils ne sont pas calcules et aucun segfault n'est genere */
void SCE_Mesh_ComputeTriangleTBN(
  float *vertex,
  float *texcoord,
  unsigned int *index,
  
  float *tangents,
  float *binormals,
  float *normals)
{
    /* vecteurs des 'cotes' de la face */
    SCE_TVector3 side0, side1;
    /* vecteurs temporaires */
    SCE_TVector3 tmpTangente, tmpNormal, tmpBinormal;
    SCE_TVector3 tmpVec1, tmpVec2;
    
    /* delta... gegnek... (valeurs bizarres pour calculs bizarres...) */
    float deltaT0, deltaT1, deltaB0, deltaB1, scale;
    
    
    /* on calcule 2 vecteurs formant les bords du triangle
       le vecteur side0 est celui allant du vertex 0 au vertex 1 */
    SCE_Vector3_Operator2v(side0, =, &vertex[index[1]*3],
                                    -, &vertex[index[0]*3]);
    
    /* le vecteur side1 est celui allant du vertex 0 au vertex 2 */
    SCE_Vector3_Operator2v(side1, =, &vertex[index[2]*3],
                                    -, &vertex[index[0]*3]);
    
    /* calculs douteux mais qui parait-il marchent...
       (et c'est verifie, ils marchent) */
    /* ici, +1 pour prendre la composante v (uv) */
    deltaT0 = texcoord[index[1]*2+1] - texcoord[index[0]*2+1];
    deltaT1 = texcoord[index[2]*2+1] - texcoord[index[0]*2+1];
    /* composante u */
    deltaB0 = texcoord[index[1]*2] - texcoord[index[0]*2];
    deltaB1 = texcoord[index[2]*2] - texcoord[index[0]*2];
    scale = 1.0 / ((deltaB0 * deltaT1) - (deltaB1 * deltaT0));
    
    
    
    if(normals != NULL)
    {
        SCE_Vector3_Cross(tmpNormal, side0, side1);
        SCE_Vector3_Normalize(tmpNormal);
        
        /* affectation */
        SCE_Vector3_Operator1v(&normals[index[0]*3], +=, tmpNormal);
        SCE_Vector3_Operator1v(&normals[index[1]*3], +=, tmpNormal);
        SCE_Vector3_Operator1v(&normals[index[2]*3], +=, tmpNormal);
    }
    
    if(tangents != NULL)
    {
        /* on calcule la tangente temporaire */
        SCE_Vector3_Operator1v(tmpVec1, = deltaT1*, side0);
        SCE_Vector3_Operator1v(tmpVec2, = deltaT0*, side1);
        SCE_Vector3_Operator1v(tmpVec1, -=, tmpVec2);
        SCE_Vector3_Operator1(tmpVec1, *=, scale);
        
        SCE_Vector3_Affectv(tmpTangente, tmpVec1);
        SCE_Vector3_Normalize(tmpTangente);
        
        /* affectation (addition) */
        SCE_Vector3_Operator1v(&tangents[index[0]*3], +=, tmpTangente);
        SCE_Vector3_Operator1v(&tangents[index[1]*3], +=, tmpTangente);
        SCE_Vector3_Operator1v(&tangents[index[2]*3], +=, tmpTangente);
    }
    
    if(binormals != NULL)
    {
        /* on calcule la binormale temporaire */
        SCE_Vector3_Operator1v(tmpVec1, = -deltaB1*, side0);
        SCE_Vector3_Operator1v(tmpVec2, = deltaB0*, side1);
        SCE_Vector3_Operator1v(tmpVec1, +=, tmpVec2);
        SCE_Vector3_Operator1(tmpVec1, *=, scale);
        
        SCE_Vector3_Affectv(tmpBinormal, tmpVec1);
        SCE_Vector3_Normalize(tmpBinormal);
        
        /* affectation (addition) */
        SCE_Vector3_Operator1v(&binormals[index[0]*3], +=, tmpBinormal);
        SCE_Vector3_Operator1v(&binormals[index[1]*3], +=, tmpBinormal);
        SCE_Vector3_Operator1v(&binormals[index[2]*3], +=, tmpBinormal);
    }
}


int SCE_Mesh_ComputeTBN(
  SCEenum polygon_type,     /* type des polygones */
  float *vertex,            /* positions des sommets */
  float *texcoord,          /* coordonnees de texture */
  unsigned int *indices,    /* indices (peut valoire NULL) */
  unsigned int icount,      /* nombre d'indices */
  unsigned int vcount,      /* nombre de sommets */
  
  float *tangents,
  float *binormals,
  float *normals)
{
    /* compteur de tours de boucle */
    unsigned int i;
    
    /* indices finaux (index si != NULL sinon i) */
    unsigned int *index = NULL;
    
    /* indices temporaires, pour les filer a ComputeTriangleTBN */
    unsigned int t_indices[3];
    
    /* count reel */
    unsigned int count;
    
    
    /* on suppose que les tableaux tangents, binormals et normals
       sont initialises a des valeurs (vecteurs) nulles (0.0;0.0;0.0) */
    
    
    /* creation d'un tableau d'indices temporaire
       si aucun tableau d'indices n'a ete donne */
    if(indices == NULL)
    {
        /* aucun index fourni, creation */
        
        count = vcount;
        
        index = SCE_malloc(sizeof *index * vcount);
        if(index == NULL)
        {
            Logger_LogSrc();
            return (SCE_ERROR);
        }
        /* tableau croissant */
        for(i=0; i<count; i++)
            index[i] = i;
    }
    else
    {
        /* des indices ont ete donnes, on les utilise */
        index = indices;
        count = icount;
    }
    
    
    /* suivant le type de polygone, on construit un triangle
       en suivant des regles bien precises, puis en donne
       notre triangle a manger a notre fonction cherie ComputeTriangleTBN */
    switch(polygon_type)
    {
        /* NOTE: penser a modifier SCECTypes.h en consequence */
        
        case SCE_TRIANGLES:
            /* la, c'est pepere */
            
            /* i+=3: pour chaque triangle */
            for(i=0; i<count; i+=3)
            {
                SCE_Mesh_ComputeTriangleTBN(vertex, texcoord, &index[i],
                                            tangents, binormals, normals);
            }
            break;
        
        case SCE_TRIANGLE_STRIP:
            /* la, ca va encore... */
            for(i=0; i<count-2; i++)
            {
                SCE_Mesh_ComputeTriangleTBN(vertex, texcoord, &index[i],
                                            tangents, binormals, normals);
            }
            break;
        
        case SCE_TRIANGLE_FAN:
            /* la, c'est deja moins rigolo... */
            
            /* tous les triangles ont un sommet en commun (le premier) donc : */
            t_indices[0] = index[0];
            
            /* en commencant par le premier sommet apres le sommet commun,
               jusqu'au dernier sommet -1 */
            for(i=1; i<count-1; i++)
            {
                t_indices[1] = index[i];
                t_indices[2] = index[i+1];
                SCE_Mesh_ComputeTriangleTBN(vertex, texcoord, t_indices,
                                            tangents, binormals, normals);
            }
            break;
        
        case SCE_QUADS:
            /* O joie, des quads... */
            
            /* i+=4: pour chaque quads */
            for(i=0; i<count; i+=4)
            {
                /* 1 quads = 2 triangles, donc : */
                
                t_indices[0] = index[i];
                t_indices[1] = index[i+1];
                t_indices[2] = index[i+2];
                SCE_Mesh_ComputeTriangleTBN(vertex, texcoord, t_indices,
                                            tangents, binormals, normals);
                
                t_indices[0] = index[i+2];
                t_indices[1] = index[i+3];
                t_indices[2] = index[i];
                SCE_Mesh_ComputeTriangleTBN(vertex, texcoord, t_indices,
                                            tangents, binormals, normals);
            }
            break;
        
        default:
            /* type de primitive non supporte, dommage pour l'utilisateur :D */
            Logger_Log(SCE_INVALID_OPERATION);
            Logger_LogMsg("primitive type unsupported, you must choose one"
                          " of the following types: SCE_TRIANGLES, "
                          "SCE_TRIANGLE_STRIP, SCE_TRIANGLE_FAN or SCE_QUADS");
            
            if(indices == NULL)
                SCE_free(index);
            return (SCE_ERROR);
            break;
    }
    
    
    if(indices == NULL)
    {
        /* si aucun index n'a ete donne, c'est donc qu'on en a cree un,
           c'est donc qu'il faut le liberer */
        SCE_free(index);
    }
    index = NULL;
    
    /* renormalisation des vecteurs */
    if(normals != NULL)
    {
        for(i=0; i<vcount; i++)
            SCE_Vector3_Normalize(&normals[i*3]);
    }
    
    if(binormals != NULL)
    {
        for(i=0; i<vcount; i++)
            SCE_Vector3_Normalize(&binormals[i*3]);
    }
    
    if(tangents != NULL)
    {
        for(i=0; i<vcount; i++)
            SCE_Vector3_Normalize(&tangents[i*3]);
    }
    
    
    return (SCE_OK);
}



/* repere les informations d'un type de sommet dans un tampon */
static int SCE_Mesh_LocateInfosBuffer(
  SCE_CVertexBuffer *buf,           /* tampon a fouiller */
  SCE_CVertexDeclaration *dec)      /* lieu de stockage des informations */
{
    unsigned int i;
    
    for(i=0; i<buf->n_decs; i++)
    {
        if(buf->dec[i].v_type == dec->v_type)
        {
            /* enregistrement des donnees */
            memcpy(dec, &buf->dec[i], sizeof *dec);
            return (SCE_OK);
        }
    }
    
    return (SCE_ERROR);
}


int SCE_Mesh_GenerateTBN(
  SCE_SMesh *mesh,          /* mesh */
  SCEuint vertid,           /* id du tampon des positions des sommets */
  SCEuint texid,            /* id du tampon des coordonnees de texture */
  int indid,                /* id du tampon des indices (peut valoir -1) */
  float *tangents,          /* tableau des tangentes */
  float *binormals,         /* tableau des binormales */
  float *normals,           /* tableau des normales */
  SCEuint texunit)          /* unite de texture comportant les
                               coordonnees de la normal-map */
{
    /* tableau des donnees */
    void *texcoordV = NULL, *vertexV = NULL, *indexV = NULL;
    float *texcoord = NULL, *vertex = NULL;
    unsigned int *index = NULL;
    GLsizei stride = 0; /* marge entre deux sommets consecutifs */
    /* counts */
    unsigned int i_count=0, v_count=0;
    
    /* pointeurs pour les conversions ... */
    GLubyte *ubyte_data = NULL;
    GLbyte *byte_data = NULL;
    GLushort *ushort_data = NULL;
    GLshort *short_data = NULL;
    GLuint *uint_data = NULL;
    GLint *int_data = NULL;
    GLfloat *float_data = NULL;
    GLdouble *double_data = NULL;
    
    /* informations sur les tableaux de donnees */
    /* type recherche: texcoord + unite de texture demandee */
    SCE_CVertexDeclaration tc_dec = {SCE_TEXCOORD0 + texunit, 0, 0, 0, 0};
    /* type recherche: positions des sommets */
    SCE_CVertexDeclaration vt_dec = {SCE_POSITION, 0, 0, 0, 0};
    
    /* compteurs */
    unsigned int i;
    int j;
    
    #ifdef __SCE_DEBUG__
    if(mesh == NULL)
    {
        Logger_Log(SCE_INVALID_ARG);
        return (SCE_ERROR);
    }
    if(texunit >= (unsigned int)SCE_CGetInteger(GL_MAX_TEXTURE_UNITS))
    {
        /* TODO: message d'erreur frequent, prevoir qq'chose... */
        Logger_Log(SCE_INVALID_ARG);
        Logger_LogMsg("the texture unit number %d "
                      "are not available in your harware", texunit);
        return (SCE_ERROR);
    }
    
    /* on verifie si les tampons de sommets demandes existent */
    if(mesh->vb[vertid].buffer == NULL)
    {
        Logger_Log(SCE_INVALID_ARG);
        Logger_LogMsg("the vertex buffer number %d is not created on this mesh",
                      vertid);
        return (SCE_ERROR);
    }
    if(mesh->vb[texid].buffer == NULL)
    {
        Logger_Log(SCE_INVALID_ARG);
        Logger_LogMsg("the vertex buffer number %d is not created on this mesh",
                      texid);
        return (SCE_ERROR);
    }
    if(indid >= 0)
    {
        if(mesh->ib[indid].buffer == NULL)
        {
            Logger_Log(SCE_INVALID_ARG);
            Logger_LogMsg("the index buffer number %d is not "
                          "created on this mesh", indid);
            return (SCE_ERROR);
        }
    }
    #endif
    
    /* on va chercher les bonnes coordonnees de texture,
       les bonnes positions de sommets ainsi que le tableau des indices */
    
    if(indid >= 0)
    {
        if((unsigned)indid < mesh->n_ib)
        {
            indexV = SCE_CLockIndexBuffer(mesh->ib[indid].buffer, GL_READ_ONLY);
            if(indexV == NULL)
            {
                Logger_LogSrc();
                return (SCE_ERROR);
            }
        }
    }
    
    /* si les donnees sont stockees dans le meme tampon,
       il ne faut pas locker celui-ci deux fois */
    if(vertid == texid)
    {
        vertexV = texcoordV =
        SCE_CLockVertexBuffer(mesh->vb[vertid].buffer, GL_READ_ONLY);
    }
    /* sinon, on verifie que les counts des deux buffers sont egaux */
    else if(mesh->vb[vertid].count == mesh->vb[texid].count)
    {
        vertexV = SCE_CLockVertexBuffer(mesh->vb[vertid].buffer, GL_READ_ONLY);
        texcoordV = SCE_CLockVertexBuffer(mesh->vb[texid].buffer, GL_READ_ONLY);
    }
    /* sinon, ils sont incompatibles et les calculs ne peuvent pas avoir lieu */
    else
    {
        Logger_Log(SCE_INVALID_OPERATION);
        Logger_LogMsg("you can't generate tangents, binormals or normals if "
                      "the buffers of datas have differents vertex count!");
        
        /* liberation des indices si besoin */
        if(indexV)
            SCE_CUnlockIndexBuffer(mesh->ib[indid].buffer);
        
        return (SCE_ERROR);
    }
    
    
    /* verification de vertexV et texcoordV, car le lockage peut echouer */
    if(vertexV == NULL || texcoordV == NULL)
    {
        Logger_LogSrc();
        
        /* liberation des tampons utilises */
        if(indexV)
            SCE_CUnlockIndexBuffer(mesh->ib[indid].buffer);
        if(vertexV)
            SCE_CUnlockVertexBuffer(mesh->vb[vertid].buffer);
        if(texcoordV)
            SCE_CUnlockVertexBuffer(mesh->vb[texid].buffer);
        
        return (SCE_ERROR);
    }
    
    
    v_count = mesh->vb[vertid].count;
    i_count = mesh->ib[indid].count;
    
    
    /* macro qui libere les tampons 'lockes' */
    #define SCE_TEMP_FREEALL()\
        if(indexV != NULL)\
            SCE_CUnlockIndexBuffer(mesh->ib[indid].buffer);\
        if(vertid == texid)\
            SCE_CUnlockVertexBuffer(mesh->vb[vertid].buffer);\
        else\
        {\
            SCE_CUnlockVertexBuffer(mesh->vb[vertid].buffer);\
            SCE_CUnlockVertexBuffer(mesh->vb[texid].buffer);\
        }
    
    if(indexV)
    {
        /* utilisation des indices */
        
        /* on alloue de la memoire si et seulement si le type
           est different de unsigned int (type des indices par defaut)
           ou le cast est inutile */
        if(mesh->ib[indid].data_type != SCE_UNSIGNED_INT)
        {
            index = SCE_malloc(sizeof *index * i_count);
            if(index == NULL)
            {
                Logger_LogSrc();
                SCE_TEMP_FREEALL();
                
                return (SCE_ERROR);
            }
        }
        
        /* "zolie" macro, dont je vais vous laisser deviner l'utilite */
        #define SCE_TRANSFERT_DATA(gltype, type)\
            case gltype:\
                type = indexV;\
                for(i=0; i<i_count; i++)\
                {\
                    index[i] = type[i];\
                }\
                type = NULL;\
                break;
        
        /* affectation des indices */
        /* recherche du type des donnees pour le cast explicite */
        switch(mesh->ib[indid].data_type)
        {
            SCE_TRANSFERT_DATA(SCE_BYTE, byte_data)
            SCE_TRANSFERT_DATA(SCE_UNSIGNED_BYTE, ubyte_data)
            SCE_TRANSFERT_DATA(SCE_SHORT, short_data)
            SCE_TRANSFERT_DATA(SCE_UNSIGNED_SHORT, ushort_data)
            SCE_TRANSFERT_DATA(SCE_INT, int_data)
            
            case SCE_UNSIGNED_INT:
                /* chouette */
                index = indexV;
                break;
            
            #ifdef __SCE_DEBUG__
            default:
                Logger_Log(SCE_INVALID_ARG);
                Logger_LogMsg("sorry, I don't support floating indices!");
                
                SCE_TEMP_FREEALL();
                SCE_free(index);
                index = NULL;
                
                return (SCE_ERROR);
            #endif
        }
        
        #undef SCE_TRANSFERT_DATA
    }
    else
    {
        /* aucun utilisation des indices */
        index = NULL;
    }
    
    
    /* recuperation des informations des donnees depuis le buffer */
    if(SCE_Mesh_LocateInfosBuffer(mesh->vb[vertid].buffer, &vt_dec) < 0)
    {
        /* la, l'utilisateur a fait une bourde */
        Logger_Log(SCE_INVALID_ARG);
        Logger_LogMsg("I can't found asked datas type on this buffer");
        
        SCE_TEMP_FREEALL();
        if(index && mesh->ib[indid].data_type != SCE_UNSIGNED_INT)
            SCE_free(index);
        index = NULL;
        
        return (SCE_ERROR);
    }
    
    /* allocation du tableau des positions des sommets */
    vertex =
    SCE_malloc(sizeof *vertex * v_count * vt_dec.n_vars);
    if(vertex == NULL)
    {
        Logger_LogSrc();
        
        SCE_TEMP_FREEALL();
        if(index && mesh->ib[indid].data_type != SCE_UNSIGNED_INT)
            SCE_free(index);
        index = NULL;
        
        return (SCE_ERROR);
    }
    
    /* "zolie" macro pour gagner en nombre de lignes de code */
    #define SCE_TRANSFERT_DATA(gltype, type)\
        case gltype:\
            stride = (vt_dec.stride > 0) ?\
                      vt_dec.stride : vt_dec.n_vars*(signed)sizeof(*type);\
            for(i=0; i<v_count; i++)\
            {\
                /* voir gl*Pointer pour plus d'infos sur vt_dec */\
                type = vertexV + vt_dec.first + (i*stride);\
                for(j=0; j<vt_dec.n_vars; j++)\
                {\
                    vertex[i*vt_dec.n_vars+j] = type[j];\
                }\
            }\
            type = NULL;\
            break;
    
    /* recherche du type des donnees pour le cast explicite */
    switch(vt_dec.type)
    {
        SCE_TRANSFERT_DATA(SCE_BYTE, byte_data)
        SCE_TRANSFERT_DATA(SCE_UNSIGNED_BYTE, ubyte_data)
        SCE_TRANSFERT_DATA(SCE_SHORT, short_data)
        SCE_TRANSFERT_DATA(SCE_UNSIGNED_SHORT, ushort_data)
        SCE_TRANSFERT_DATA(SCE_INT, int_data)
        SCE_TRANSFERT_DATA(SCE_UNSIGNED_INT, uint_data)
        SCE_TRANSFERT_DATA(SCE_FLOAT, float_data)
        SCE_TRANSFERT_DATA(SCE_DOUBLE, double_data)
        
        #undef SCE_TRANSFERT_DATA
        
        #ifdef __SCE_DEBUG__
        default:
            Logger_Log(SCE_INVALID_ARG);
            Logger_LogMsg("DEBUG: an error was detected here! bad type");
            
            SCE_TEMP_FREEALL();
            
            if(index && mesh->ib[indid].data_type != SCE_UNSIGNED_INT)
                SCE_free(index);
            index = NULL;
            
            SCE_free(vertex);
            vertex = NULL;
            
            return (SCE_ERROR);
        #endif
    }
    
    
    /* recuperation des informations sur les coordonnees de texture */
    if(SCE_Mesh_LocateInfosBuffer(mesh->vb[texid].buffer, &tc_dec) < 0)
    {
        /* la, l'utilisateur a fait une bourde */
        Logger_Log(SCE_INVALID_ARG);
        Logger_LogMsg("I can't found asked datas type on this buffer");
        
        SCE_TEMP_FREEALL();
        if(index && mesh->ib[indid].data_type != SCE_UNSIGNED_INT)
            SCE_free(index);
        index = NULL;
        
        SCE_free(vertex);
        vertex = NULL;
        
        return (SCE_ERROR);
    }
    
    /* allocation du tableau des coordonnees de texture */
    texcoord =
    SCE_malloc(sizeof *vertex * v_count * tc_dec.n_vars);
    if(texcoord == NULL)
    {
        Logger_LogSrc();
        
        SCE_TEMP_FREEALL();
        if(index && mesh->ib[indid].data_type != SCE_UNSIGNED_INT)
            SCE_free(index);
        index = NULL;
        
        SCE_free(vertex);
        vertex = NULL;
        
        return (SCE_ERROR);
    }
    
    
    /* "zolie" macro pour gagner en nombre de lignes de code */
    #define SCE_TRANSFERT_DATA(gltype, type)\
        case gltype:\
            stride = (tc_dec.stride > 0) ?\
                      tc_dec.stride : tc_dec.n_vars*(signed)sizeof(*type);\
            for(i=0; i<v_count; i++)\
            {\
                /* voir gl*Pointer pour plus d'infos sur tc_dec */\
                type = texcoordV + tc_dec.first + (i*stride);\
                for(j=0; j<tc_dec.n_vars; j++)\
                {\
                    texcoord[i*tc_dec.n_vars+j] = type[j];\
                }\
            }\
            type = NULL;\
            break;
    
    /* recherche du type des donnees pour le cast explicite */
    switch(tc_dec.type)
    {
        SCE_TRANSFERT_DATA(SCE_BYTE, byte_data)
        SCE_TRANSFERT_DATA(SCE_UNSIGNED_BYTE, ubyte_data)
        SCE_TRANSFERT_DATA(SCE_SHORT, short_data)
        SCE_TRANSFERT_DATA(SCE_UNSIGNED_SHORT, ushort_data)
        SCE_TRANSFERT_DATA(SCE_INT, int_data)
        SCE_TRANSFERT_DATA(SCE_UNSIGNED_INT, uint_data)
        SCE_TRANSFERT_DATA(GL_FLOAT, float_data)
        SCE_TRANSFERT_DATA(GL_DOUBLE, double_data)
        
        #undef SCE_TRANSFERT_DATA
        
        #ifdef __SCE_DEBUG__
        default:
            Logger_Log(SCE_INVALID_ARG);
            Logger_LogMsg("DEBUG: an error was detected here! bad type");
            
            SCE_TEMP_FREEALL();
            
            if(index && mesh->ib[indid].data_type != SCE_UNSIGNED_INT)
                SCE_free(index);
            index = NULL;
            
            SCE_free(vertex);
            vertex = NULL;
            
            SCE_free(texcoord);
            texcoord = NULL;
            
            return (SCE_ERROR);
        #endif
    }
    
    
    SCE_Mesh_ComputeTBN(mesh->polygon_type,
                        vertex, texcoord, index,
                        i_count, v_count,
                        tangents, binormals, normals);
    
    SCE_TEMP_FREEALL();
    
    #undef SCE_TEMP_FREEALL
    
    if(index && mesh->ib[indid].data_type != SCE_UNSIGNED_INT)
        SCE_free(index);
    index = NULL;
    
    SCE_free(vertex);
    vertex = NULL;
    
    SCE_free(texcoord);
    texcoord = NULL;
    
    return (SCE_OK);
}


/* ajoute le 25/06/2007 */
/* revise le 27/06/2007 */
int SCE_Mesh_AddGenVertices(
  SCE_SMesh *mesh,
  SCEenum gen_type,
  SCEuint vertid,            /* id du tampon des positions des sommets */
  SCEuint texid,             /* id du tampon des coordonnees de texture */
  int indid,                 /* id du tampon des indices (peut valoir -1) */
  SCEuint buf,
  SCEuint attrib_type,
  SCEuint texunit)
{
    float *t = NULL, *b = NULL, *n = NULL;
    unsigned int i;
    size_t nvars;
    
    #ifdef __SCE_DEBUG__
    if(mesh == NULL)
    {
        Logger_Log(SCE_INVALID_ARG);
        return (SCE_ERROR);
    }
    #endif
    
    nvars = mesh->vb[vertid].count * 3;
    
    #define SCE_MESH_DEL()\
    {\
        Logger_LogSrc();\
        SCE_free(n);\
        SCE_free(b);\
        SCE_free(t);\
        return (SCE_ERROR);\
    }
    
    /* analyse des types a generer demandes */
    if(gen_type & SCE_GEN_TANGENTS)
    {
        t = SCE_malloc(nvars * sizeof *t);
        if(t == NULL)
            SCE_MESH_DEL()
        for(i=0; i<nvars; i++)
            t[i] = 0.0f;
    }
    if(gen_type & SCE_GEN_BINORMALS)
    {
        b = SCE_malloc(nvars * sizeof *b);
        if(b == NULL)
            SCE_MESH_DEL()
        for(i=0; i<nvars; i++)
            b[i] = 0.0f;
    }
    if(gen_type & SCE_GEN_NORMALS)
    {
        n = SCE_malloc(nvars * sizeof *n);
        if(n == NULL)
            SCE_MESH_DEL()
        for(i=0; i<nvars; i++)
            n[i] = 0.0f;
    }
    /***/
    
    
    /* generation */
    SCE_Mesh_GenerateTBN(mesh, vertid, texid, indid, t, b, n, texunit);
    
    
    /** TODO: attrib_type peut devenir trop grand */
    /* ajout */
    if(gen_type & SCE_GEN_TANGENTS)
    {
        if(SCE_Mesh_AddVertices(mesh, buf, attrib_type, SCE_FLOAT,
                                3, nvars, t, SCE_TRUE) < 0)
            SCE_MESH_DEL()
        attrib_type++;
    }
    if(gen_type & SCE_GEN_BINORMALS)
    {
        if(SCE_Mesh_AddVertices(mesh, buf, attrib_type, SCE_FLOAT,
                                3, nvars, b, SCE_TRUE) < 0)
            SCE_MESH_DEL()
        attrib_type++;
    }
    if(gen_type & SCE_GEN_NORMALS)
    {
        if(SCE_Mesh_AddVertices(mesh, buf, attrib_type, SCE_FLOAT,
                                3, nvars, n, SCE_TRUE) < 0)
            SCE_MESH_DEL()
    }
    /***/
    
    #undef SCE_MESH_DEL
    
    return (SCE_OK);
}



int SCE_Mesh_Build(SCE_SMesh *mesh)
{
    /* compteur */
    unsigned int i;
    
    #ifdef __SCE_DEBUG__
    if(mesh == NULL)
    {
        Logger_Log(SCE_INVALID_ARG);
        return (SCE_ERROR);
    }
    #endif
    
    
    for(i=0; i<SCE_MAX_BUFFERS; i++)
    {
        if(!mesh->vb[i].builded)
        {
            SCE_CBuildVertexBuffer(mesh->vb[i].buffer, mesh->vb[i].usage);
            mesh->vb[i].builded = 1;
        }
        
        if(!mesh->ib[i].builded)
        {
            SCE_CBuildIndexBuffer(mesh->ib[i].buffer, mesh->ib[i].usage);
            mesh->ib[i].builded = 1;
        }
    }
    
    mesh->builded = 1;
    
    return (SCE_OK);
}


void SCE_Mesh_SetRenderMode(SCE_SMesh *mesh, SCEenum mode)
{
    if(mesh != NULL)
        mesh->polygon_type = mode;
}


int SCE_Mesh_Update(SCE_SMesh *mesh, unsigned int bid)
{
    #ifdef __SCE_DEBUG__
    if(mesh == NULL)
    {
        Logger_Log(SCE_INVALID_ARG);
        return (SCE_ERROR);
    }
    /* NOTE: devient errone si le nombre de tampon passe en illimite */
    if(bid >= SCE_MAX_BUFFERS)
    {
        Logger_Log(SCE_INVALID_ARG);
        return (SCE_ERROR);
    }
    /* inutile dans la mesure ou CUpdateVertexBuffer supporte NULL
       mais pour le debugging, ca peut etre important de recuperer ces infos */
    if(mesh->vb[bid].buffer == NULL)
    {
        Logger_Log(SCE_INVALID_OPERATION);
        Logger_LogMsg("you can't update a uninitialized buffer");
        return (SCE_ERROR);
    }
    #endif
    
    if(SCE_CUpdateVertexBuffer(mesh->vb[bid].buffer) < 0)
    {
        Logger_LogSrc();
        return (SCE_ERROR);
    }
    
    return (SCE_OK);
}



void SCE_Mesh_Bind(SCE_SMesh *mesh)
{
    unsigned int i;
    
    m = mesh;
    
    if(m)
    {
        /* pour chaque tampon... */ /* NOTE: devient errone si on sait quoi.. */
        for(i=0; i<SCE_MAX_BUFFERS; i++)
        {
            /* ...on l'active s'il a ete specifie comme actif pour le rendu */
            if(m->vb_actives[i])
                SCE_CUseVertexBuffer(m->vb[i].buffer);
        }
        
        if(m->use_indices)
            SCE_CUseIndexBuffer(m->ib[m->ib_active].buffer);
    }
    else
    {
        SCE_CUseIndexBuffer(NULL);
        SCE_CUseVertexBuffer(NULL);
    }
}


/* revise le 14/06/2007 */
void SCE_Mesh_Draw(void)
{
    if(m->use_indices)
        SCE_CDrawIndexedBuffer(m->polygon_type, 0,
                               m->ib[m->ib_active].data_type,
                               m->active_icount);
    else
        SCE_CDrawBuffer(m->polygon_type, 0, m->active_vcount);
}


void
SCE_Mesh_Render(
  unsigned int first,
  unsigned int nb,
  SCE_SMesh **mesh)
{
    unsigned int i, j;
    
    for(i=first; i<nb; i++)
    {
        /* pour chaque tampon... */ /* NOTE: devient errone si on sait quoi.. */
        for(j=0; j<SCE_MAX_BUFFERS; j++)
        {
            /* ...on l'active s'il a ete specifie comme actif pour le rendu */
            if(mesh[i]->vb_actives[j])
                SCE_CUseVertexBuffer(mesh[i]->vb[j].buffer);
        }
        
        if(mesh[i]->use_indices)
        {
            SCE_CUseIndexBuffer(mesh[i]->ib[mesh[i]->ib_active].buffer);
            SCE_CDrawIndexedBuffer(mesh[i]->polygon_type, 0,
                                   mesh[i]->ib[mesh[i]->ib_active].data_type,
                                   mesh[i]->active_icount);
        }
        else
        {
            SCE_CDrawBuffer(mesh[i]->polygon_type, 0, mesh[i]->active_vcount);
        }
    }
    
    SCE_CUseIndexBuffer(NULL);
    SCE_CUseVertexBuffer(NULL);
}
