/***************************************************************************
 *   Copyright (C) 2007  Antony Martin <martin.antony@yahoo.fr>            *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or         *
 *   modify it under the terms of the GNU General Public License           *
 *   as published by the Free Software Foundation; either version 2        *
 *   of the License, or (at your option) any later version.                *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA          *
 ***************************************************************************/

/* Cree le : 26 fevrier 2007
   derniere modification le 15/05/2007 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <SCE/utils/SCEString.h>

#include <SCE/utils/SCETime.h>


/* retourne la chaine du mois nb */
static const char* SCE_Time_GetMonth(const int nb)
{
    /* chaines statiques des noms des mois de l'annee ranges
     * dans l'ordre croissant. (janvier a decembre)
     * - 12 mois
     * - 4 caracteres (caractere de fin de ligne compris)
     */
    static const char months[12][4] =
    {
        "Jan","Feb","Mar","Apr","May","Jun","Jul","Aug","Sep","Oct","Nov","Dec"
    };
    
    return (months[nb]);
}

/* retourne la chaine du jour de la semaine nb */
static const char* SCE_Time_GetDay(const int nb)
{
    /* chaines statiques des noms des jours de la semaine ranges
     * dans l'ordre croissant. (lundi a dimanche)
     * - 7 jours
     * - 4 caracteres (caractere de fin de ligne compris)
     */
    static const char days[7][4] =
    {
        "Sun","Mon","Tue","Wed","Thu","Fri","Sat"
    };
    
    return (days[nb]);
}


void SCE_Time_MakeString(char *str, const struct tm* const info)
{
    char s_hour[9];
    char *stmp = NULL;
    
    #ifdef __SCE_DEBUG__
    if(!str || !info)
    {
        return;
    }
    #endif
    
    /* initialisation */
    memset(s_hour, '\0', sizeof s_hour);
    
    /* conversion de l'heure.
     * n'utilise pas sprintf car sinon on risque de tomber sur :
     * 20:5:6 au lieu de : 20:05:06
     */
    if(info->tm_hour <= 9)
        strcat(s_hour, "0");
    
    stmp = SCE_String_Strof(info->tm_hour, 0);
    strcat(s_hour, stmp);
    free(stmp);
    stmp = NULL;
    strcat(s_hour, ":");
    
    if(info->tm_min <= 9)
        strcat(s_hour, "0");
    
    stmp = SCE_String_Strof(info->tm_min, 0);
    strcat(s_hour, stmp);
    free(stmp);
    stmp = NULL;
    strcat(s_hour, ":");
    
    if(info->tm_sec <= 9)
        strcat(s_hour, "0");
    
    stmp = SCE_String_Strof(info->tm_sec, 0);
    strcat(s_hour, stmp);
    free(stmp);
    stmp = NULL;
    
    /* assemblage */
    sprintf(str, "%s %s %d %s %d",
            SCE_Time_GetDay(info->tm_wday),
            SCE_Time_GetMonth(info->tm_mon),
            info->tm_mday,
            s_hour, /* TODO: il manque le decalage horaire */
            /* voir -man 3 gmtime */
            info->tm_year+1900);
}
